import numpy as np

from . import LOG


def compute_q(p_w: float, cos_phi: float, mode: str = "inductive") -> float:
    """Calculates reactive power

    Reactive power is calculated with `p_w * tan(arccos(cos_phi))`.
    If mode equals `capacitive`, the returned value is negative.

    Parameters
    ----------
    p_w : float
        The active power (can also be kW)
    cos_phi : float
        The phase angle to calculate q.
    mode : str, optional
        Can be either 'inductive' or 'capacitive'. Defaults to
        'inductive', which returns the value as it is. If set to
        'capacitive', the sign of the output is flipped.

    Returns
    -------
    float
        Returns *q_var* in the same size of order like p_w (e.g., if
        *p* is in kW, *q* will be in kvar)

    """
    abs_q = p_w * np.tan(np.arccos(cos_phi))
    # inductive load 'consumes' reactive power
    if mode == "inductive":
        return abs_q

    # capacitve load 'provides' reactive power
    elif mode == "capacitve":
        return -1 * abs_q
    else:
        LOG.warning(
            "Illegal mode: %s. Falling back to default (inductive).",
            str(mode),
        )
        return abs_q



"""This module contains the model for data simulators."""

import math

import numpy as np
from midas.util.compute_q import compute_q


class DataModel:
    """A model for a single load or sgen time series.

    This model can be used to "simulate" a time series of active and,
    if provided, reactive power values. The model is designed to
    simulate one year but the time resolution of the data needs to be
    specified. This time resolution can be different than the step size
    for one step but if the step size is lower, values will be returned
    repeatedly.

    If that behavior is not desired, the linear interpolation function
    of the model can be used. This can also be combined with the
    randomization function. Those can be activated by passing the
    corresponding flags (see Parameter section).

    If the data does not contain a reactive power time series, the
    model will calculate reactive power based on the cos phi, which
    should be provided before each step. There is an option to
    randomize the output of the cos phi calculation as well.

    The main input for the model is a datetime object for the time to
    be simulated. Month, day, hour, minute, and second will be used to
    determine the corresponding value of the data set. Year information
    will be dropped.

    This two consequences. First, the model can be used to simulate
    several years and, secondly, the data set needs to be exactly for
    one year. Larger data sets will be "cropped" to one year and
    smaller data sets will probably lead to an error.

    Parameters
    ----------
    data_p : pandas.DataFrame
        Contains values (either load or sgen) for active power. The
        index columns are simple *int*.
    data_q : pandas.DataFrame
        Contains values (either load or sgen) for reactive power. The
        index columns are simple *int*. If *None* is provided for
        *data_q*, the cos phi is used in each step to calculate a value
        for *q*.
    data_step_size: int
        Timely resolution of the data in seconds.
    scaling : float
        An overall scaling for this model. This scaling is applied in
        each to step both of *p* and *q*.
    seed : int, optional
        A seed for the random number generator.
    interpolate : bool, optional
        If set to *True*, interpolation is applied when values between
        full 15-minute intervalls are requested.
    randomize_data : bool, optional
        If set to *True*, a normally distributed random noise is
        applied to all outputs.
    randomize_cos_phi : bool, optional
        If set to *True* and data_q is not provided, the cos phi for
        the calculation of *q* is randomized.
    date_index : bool, optional
        Set this to *True* if the data has datetime as index instead of
        ints (planned but not yet supported).
    noise_factor: float, optional
        Set this to increase or lower the noise if randomization is
        activated.

    Attributes
    ----------
    now_dt : datetime.datetime
        *now_dt is an input and needs to be provided in each step.*
        The current simulated time. Is used to calculate the index for
        the current value in the time series.
    cos_phi : float
        *cos_phi is an input and needs to be provided in each step.*
        The phase angle used to calculate reactive power if no reactive
        power time series is provided.
    p_mw : float
        *p_mw is an output.* The active power for the current step.
    q_mvar : float
        *q_mvar is an output.* The reactive power for the current step.

    """

    def __init__(self, data_p, data_q, data_step_size, scaling, **params):
        self.data_p = data_p
        self.data_q = data_q
        self.sps = data_step_size
        self.scaling = scaling
        self.date_index = params.get("date_index", False)

        # RNG
        self.seed = params.get("seed", None)
        self.rng = np.random.RandomState(self.seed)

        self.interpolate = params.get("interpolate", False)
        self.randomize_data = params.get("randomize_data", False)
        self.randomize_cos_phi = params.get("randomize_cos_phi", False)
        self.noise_factor = params.get("noise_factor", 0.2)
        self.p_std = self.data_p.std() * self.scaling
        self.p_mwh_per_a = self.data_p.sum() * self.scaling / self.sps * 3_600
        if self.data_q is not None:
            self.q_std = self.data_q.std() * self.scaling
        else:
            self.q_std = None

        # Inputs
        self.now_dt = None
        self.cos_phi = None

        # Outputs
        self.p_mw = None
        self.q_mvar = None

    def step(self):
        """Perform a simulation step."""

        self.p_mw = None
        self.q_mvar = None

        self._interpolation()
        self._randomization()

        self._random_cos_phi()

        if self.q_mvar is None:
            self.q_mvar = compute_q(self.p_mw, self.cos_phi)

        # Numpy multiply failes for int and float if *= is used
        self.p_mw = self.p_mw * self.scaling
        self.q_mvar = self.q_mvar * self.scaling

        tmp = self.p_mw**2 + self.q_mvar**2
        if tmp != 0:
            self.cos_phi = self.p_mw / math.sqrt(tmp)
        else:
            self.cos_phi = 0

    def _interpolation(self):
        # We assume that the dataset starts on the first of the year.
        dif = self.now_dt - self.now_dt.replace(
            month=1, day=1, hour=0, minute=0, second=0
        )
        dif_s = dif.total_seconds()

        tidx = int(dif_s // self.sps) % len(
            self.data_p.index
        )  # step size of the data

        # Apply interpolation
        if self.interpolate:
            self.p_mw = self._interpolate(tidx, self.data_p)
            if self.data_q is not None:
                self.q_mvar = self._interpolate(tidx, self.data_q)
        else:
            self.p_mw = self.data_p.values[tidx]
            if self.data_q is not None:
                self.q_mvar = self.data_q.values[tidx]

    def _interpolate(self, tidx, data):

        current_second = (
            self.now_dt.minute * 60 + self.now_dt.second
        ) % self.sps

        tidx_end = tidx + 2
        y_vals = data.values[tidx:tidx_end]

        if tidx_end > len(data.index):
            # Start from beginning
            y_vals = np.array([y_vals[0], data.values[0]])
        x_vals = np.array([0, self.sps])

        result = np.interp(current_second, x_vals, y_vals)

        return result

    def _randomization(self):

        if self.randomize_data:
            noise = self.rng.normal(
                scale=(self.p_std * self.noise_factor), loc=0.0
            )
            self.p_mw = max(0, self.p_mw + noise)

            if self.q_mvar is not None:
                noise = self.rng.normal(
                    scale=(self.q_std * self.noise_factor), loc=0.0
                )
                self.q_mvar = max(0, self.q_mvar + noise)

    def _random_cos_phi(self):
        if self.q_mvar is None and self.randomize_cos_phi:
            self.cos_phi = max(
                0, min(1.0, self.rng.normal(scale=0.02, loc=0.9))
            )



import pprint
from typing import Any

from .runtime_config import RuntimeConfig


def mformat(msg: Any) -> str:

    if RuntimeConfig().misc["use_pprint"]:
        return pprint.pformat(msg)
    else:
        return str(msg)



import os
import subprocess

from . import LOG


def convert_markdown(report_file_path: str) -> bool:
    """Try to convert the markdown file to pdf or odt."""

    success = False
    if not os.path.isfile(report_file_path):
        LOG.warning(f"{report_file_path} is not a file. Aborting!")
        return success

    pdf_available = True
    try:
        subprocess.check_output(["pandoc", "--version"])
    except FileNotFoundError:
        # No pandoc available, therefore nothing else to do here.
        return False

    try:
        subprocess.check_output(
            [
                "pandoc",
                "--template",
                "eisvogel",
                "--listings",
                "-o",
                f"{report_file_path[:-3]}.pdf",
                report_file_path,
            ]
        )
        return True
    except FileNotFoundError as err:
        # Probably the template is not available
        LOG.debug(
            "Trying to use pandoc with template resulted in following error: "
            "%s",
            err,
        )

    except subprocess.CalledProcessError as err:
        # Probably no pdf engine installed
        LOG.debug("Trying to use pandoc resulted in following error: %s", err)
        pdf_available = False

    if pdf_available:
        try:
            subprocess.check_output(
                [
                    "pandoc",
                    "--listings",
                    "-o",
                    f"{report_file_path[:-3]}.pdf",
                    report_file_path,
                ]
            )
            return True
        except Exception as err:
            LOG.debug(
                "Trying to use pandoc resulted in following error: %s", err
            )

    try:
        subprocess.check_output(
            [
                "pandoc",
                "-s",
                "-t",
                "odt",
                "-o",
                f"{report_file_path[:-3]}.odt",
                report_file_path,
            ]
        )
        return True
    except Exception:
        pass

    LOG.warning(
        "Unable to convert markdown file. Check debug logs for more info."
    )
    return False



"""MIDAS base upgrade module for dataset simulators."""
import logging

from .upgrade_module import UpgradeModule

LOG = logging.getLogger(__name__)


class BaseDataModule(UpgradeModule):
    def __init__(
        self,
        module_name,
        default_scope_name,
        default_sim_config_name,
        default_import_str,
        default_cmd_str,
        log=LOG,
    ):
        super().__init__(
            module_name=module_name,
            default_scope_name=default_scope_name,
            default_sim_config_name=default_sim_config_name,
            default_import_str=default_import_str,
            default_cmd_str=default_cmd_str,
            log=log,
        )

        self.models = {}
        self._scaling_key = "scaling"

    def check_module_params(self, module_params):
        """Check the module params and provide default values"""

        module_params.setdefault("start_date", self.scenario.base.start_date)
        module_params.setdefault("data_path", self.scenario.base.data_path)
        module_params.setdefault("cos_phi", self.scenario.base.cos_phi)
        module_params.setdefault("interpolate", False)
        module_params.setdefault("noise_factor", 0.2)

        if module_params["no_rng"]:
            module_params["randomize_data"] = False
            module_params["randomize_cos_phi"] = False
        else:
            module_params.setdefault("randomize_data", False)
            module_params.setdefault("randomize_cos_phi", False)

    def check_sim_params(self, module_params):
        """Check the params for a certain scope/simulator instance."""

        self.sim_params.setdefault("grid_name", self.scope_name)
        self.sim_params.setdefault("start_date", module_params["start_date"])
        self.sim_params.setdefault("data_path", module_params["data_path"])
        self.sim_params.setdefault("cos_phi", module_params["cos_phi"])
        self.sim_params.setdefault("interpolate", module_params["interpolate"])
        self.sim_params.setdefault(
            "randomize_data", module_params["randomize_data"]
        )
        self.sim_params.setdefault(
            "randomize_cos_phi", module_params["randomize_cos_phi"]
        )
        self.sim_params.setdefault(
            "noise_factor", module_params["noise_factor"]
        )
        self.sim_params.setdefault("seed_max", self.scenario.base.seed_max)
        self.sim_params.setdefault("seed", self.scenario.create_seed())

    def start_models(self):
        """Start models of a certain simulator."""

        for model in self.models:
            mapping_key = f"{model}_mapping"

            self.start_models_from_mapping(mapping_key, model, "load")

    def start_models_from_mapping(self, mapping_key, model, mtype):
        """Iterate over a mapping and start models."""

        self.sim_params.setdefault(mapping_key, self.create_default_mapping())
        if not self.sim_params[mapping_key]:
            # No mappings configure
            return

        mapping = self.scenario.create_shared_mapping(
            self, self.sim_params["grid_name"], mtype
        )
        model_name = model
        for bus, entities in self.sim_params[mapping_key].items():
            mapping.setdefault(bus, [])
            for eidx, (name, scale) in enumerate(entities):
                if model is None:
                    model_name = name

                model_key = self.scenario.generate_model_key(
                    self, model_name, bus, eidx
                )
                scaling = scale * float(self.sim_params[f"{mtype}_scaling"])
                params = {self._scaling_key: scaling}
                full_id = self.start_model(model_key, model_name, params)

                info = self.scenario.get_sim(self.sim_key).get_data_info()
                mapping[bus].append(
                    (
                        model_name,
                        info[full_id.split(".")[-1]]["p_mwh_per_a"] * scaling,
                    )
                )

    def connect_to_grid(self, mapping_key, model, mtype, attrs=None):
        if attrs is None:
            attrs = ["p_mw", "q_mvar"]

        model_name = model
        for bus, entities in self.sim_params[mapping_key].items():
            for eidx, (name, _) in enumerate(entities):
                if model is None:
                    model_name = name

                model_key = self.scenario.generate_model_key(
                    self, model_name, bus, eidx
                )
                grid_entity_key = self.get_grid_entity(mtype, bus)
                self.connect_entities(model_key, grid_entity_key, attrs)

    def connect_to_db(self):
        """Connect the models to db."""
        for model in self.models:
            mapping_key = f"{model}_mapping"
            self._connect_to_db(mapping_key, model)

    def _connect_to_db(self, mapping_key, model, attrs=None):
        if attrs is None:
            attrs = ["p_mw", "q_mvar"]

        db_key = self.scenario.find_first_model("store", "database")[0]
        model_name = model
        for bus, entities in self.sim_params[mapping_key].items():
            for eidx, (name, _) in enumerate(entities):
                if model is None:
                    model_name = name

                model_key = self.scenario.generate_model_key(
                    self, model_name, bus, eidx
                )

                self.connect_entities(model_key, db_key, attrs)

    def get_grid_entity(self, mtype, bus):
        models = self.scenario.find_grid_entities(
            self.sim_params["grid_name"], mtype, endswith=f"_{bus}"
        )
        if models:
            for key in models:
                # Return first match
                return key

        self.logger.info(
            "Grid entity for %s, %s at bus %d not found",
            self.sim_params["grid_name"],
            mtype,
            bus,
        )
        raise ValueError(
            f"Grid entity for {self.sim_params['grid_name']}, {mtype} "
            f"at bus {bus} not found!"
        )

    def create_default_mapping(self):
        default_mapping = dict()

        return default_mapping
